/** 
 * SALSA 2
 * July 2000
 * Mike Gasser (gasser@indiana.edu)
 */

// INCLUDE STATEMENTS

import java.awt.Polygon;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Font;

/**
* A single hexagonal space within the World which keeps tracks of its
* contents (a Thing) and its 6 neighbors (other Cells).  Knows how to
* paint itself in various ways.  All of these get instantiated when World
* is instantiated.
*/

public class Cell {

  // CONSTANTS
    
  public final static Color
    FERTILITY_COLOR = Color.darkGray,
    LIGHTEST_GRAY = new Color(225, 225, 225),
    BG_COLOR = Color.black,
    // Some new colors
    LIGHT_GREEN = new Color(160, 255, 160),
    LIGHT_BLUE = new Color(160, 160, 255),
    MEDIUM_BLUE = new Color(80, 80, 255);

  public final static double
    SIXTH_PI = Math.PI / 6.0;
    
  public final static Font
    FONT = new Font("Monospaced", Font.PLAIN, 9);
  
  public final static int
    // The 6 cardinal directions.
    N = 0,
    NE = 1,
    SE = 2,
    S = 3,
    SW = 4,
    NW = 5,
    
    // Represent the color of different Things that a Cell
    // may contain; used by Vision.
    BLACK = 0,
    BLUE = 1,
    GREEN = 2,
    RED = 3,
    MAGENTA = 4,
    YELLOW = 5,
    CYAN = 6,
    GRAY = 7,
    DARK_GRAY = 8,
    LIGHT_GRAY = 9,
    PINK = 10,
    LGREEN = 11,
    LBLUE = 12,
    
    EMPTY_COLOR = 8,
    
    // Represent the "feel" of different Things that a Cell
    // may contain; used by Touch.
    EMPTY = 0,
    HARD = 1,    // Rock
    SOFT = 2,    // tender Plant
    ROUGH = 3,
    SQUISHY = 4; // Critter
  
  // Names of the different "feels".  Positions are those of
  // the ints for each feel. Useful for debugging.
  public static final String[]
    FEEL_NAMES =
    {"Empty", "Hard", "Soft", "Rough", "Squishy"};
    
  // Names of the different colors.  Positions are those of
  // the ints for each color.  Useful for debugging.
  public static final String[]
    COLOR_NAMES =
    {"Black", "Blue", "Green", "Red", "Magenta", "Yellow",
    "Cyan", "Gray", "Dark Gray", "Light Gray",
    "Pink", "LightGreen", "LightBlue"};
    
  // The possible colors.  Positions are those of the ints
  // for each color (above).
  public static final Color[]
    COLORS = 
    {Color.black, Color.blue, Color.green,
     Color.red, Color.magenta, Color.yellow,
     Color.cyan, Color.gray, Color.darkGray, Color.lightGray,
     Color.pink, LIGHT_GREEN, LIGHT_BLUE};

  // Dimensions of the center dot showing and empty Cell position.
  private final static int
    GRID_DOT_RADIUS = 1,
    GRID_DOT_DIAMETER = 3;
  
  // VARIABLES
  
  private int
    // Pixel coordinates of the upper-left corner
    x, y,
    // Cell coordinates
    column, row,
    // "Radius" of the hexagon
    radius,
    // Center of the hexagon
    centerX, centerY,
    // Where character is written
    xText, yText,
    // Radius and diameter of circle to be used to fill Cell
    circleRad, circleDiam,
    smallCircleRad, smallCircleDiam,
    // Points on the circle for each section
    circleX[], circleY[],
    // X and Y coordinates of the vertices of the hexagon
    hexX[], hexY[];
  
  // Helps to have direct access to the WorldCanvas where Cells
  // are drawn.
  public WorldCanvas canvas;
  
  // The UI where messages are written.
  private UI ui;
    
  // The hexagon that defines the way the Cell is drawn.
  private Polygon
    hex = null;
  
  // The Thing that is in the Cell, null if the Cell is empty.
  private Thing
    contents = null;

  // The 6 neighbors of the Cell, accessed using the cardinal
  // direction constants.
  private Cell
    neighbors[];
    
  private Color
    bgColor = null,
    outlineColor = null;
    
  private boolean
    fertile = false;
  
  // CONSTRUCTOR
  
  /**
  *  When Cell is instantiated, it initializes its variables,
  *  including hex, but does not set its neighbors.  This happens
  *  in World.   
  */
  Cell (UI ui, int column, int row, int x, int y, int radius) {
    this.x = x;
    this.y = y;
    this.column = column;
    this.row = row;
    this.radius = radius;
    this.ui = ui;
    neighbors = new Cell[6];
    makeHex();
    setCircleValues();
  } // constructor Cell
  
  // METHODS
 
  // Methods that access variables.
  
  /**
  * Returns the current contents of the Cell.
  */
  public Thing getContents() {
      return contents;
  } // method getContents
  
  private int getFeel() {
      if (contents == null)
	  return EMPTY;
      else if (contents instanceof Rock)
	  return HARD;
      else if (contents instanceof Critter)
	  return SQUISHY;
      else if (((Plant)contents).isTough())
	  return ROUGH;
      else
	  return SOFT;
  } // getFeel()
  
  /**
  * Returns the "feel" of the Thing in the next Cell in the
  * given direction (assumes there is such a Cell).  Used by Touch.
  */
  public int getNextFeel(int direction) {
      return (getNeighbor(direction).getFeel());
  } // method getNextFeel
  
  /**
  * Returns the integer for the color of whatever is in this
  * Cell.
  */
  public int getColorInt() {
      if (contents == null)
	  return BLACK;
      else if (contents instanceof Rock)
	  return BLUE;
      else if (contents instanceof Plant)
	  return GREEN;
      else if (contents instanceof Critter)
	  return RED;
      else
	  return BLACK;
  } // method getColorInt()
  
  public Color getColor() {
      if (contents == null)
	  return COLORS[EMPTY_COLOR];
      else
	  return contents.getColor();
  } // method getColor()

  /**
  * Returns the neighboring Cell in the given cardinal direction.
  */
  public Cell getNeighbor(int direction) {
      return (neighbors[direction]);
  } // getNeighbor(int)
  
  public Cell[] getNeighbors() {
      return neighbors;
  } // getNeighbors()
  
  /**
  * Return the pixel coordinates of the center of
  * the Cell.
  */
  public int getCenterX() {
      return centerX;
  } // method getCenterX
  
  public int getCenterY() {
      return centerY;
  } // method getCenterY
  
  /**
  * A name for the Cell.
  */
  public String getName() {
      return ("Cell " + x + "/" + y);
  } // method getName
  
  public Graphics getGraphics() {
      return canvas.getGraphics();
  } // getGraphics()
  
  /**
  * Is the Cell empty of Things?
  */
  public boolean isEmpty() {
    return (contents == null);
  } // method isEmpty
  
  /**
  * Is the point (x,y) inside the Cell's hexagon?
  */
  public boolean contains(int x, int y) {
      return (hex.contains(x, y));
  } // method contains
  
  /**
  * Returns a string representing the contents of the Cell.
  */
  public String getContentsAsString() {
      if (contents == null)
        return ("Empty");
      else
        return (contents.getName());
  } // method getContentsAsString

  /**
  * Returns the Cell's column.
  */
  public int getColumn () {
      return column;
  } // method getColumn
  
  /**
  * Returns the Cell's row.
  */
  public int getRow () {
      return row;
  } // method getRow
  
  // Methods relating to neighboring Cells.
  
  /**
  * Does the Cell have any neighbors that are filled, other
  * than by the given Thing?
  */
  public boolean anyNeighborsFilled(Thing exceptThing) {
      boolean filled = false;
      int direction = 0;
      Cell neighbor;
      Thing contents;
      
      while (direction < 6 && !filled) {
        neighbor = getNeighbor (direction);
        if (neighbor != null) {
            contents = neighbor.getContents();
            if (contents != null && contents != exceptThing)
	      filled = true;
	direction++;
        }
      }
      return filled;
  } // method anyNeighborsFilled

  /**
  * Returns the first randomly selected neighboring Cell that
  * is found (null if none).  If Params.thingsCanAbut is false,
  * looks for a Cell which doesn't border on anything other
  * than Cells in exceptThing.
  */
  public Cell getRandomEmptyNeighbor(Thing exceptThing) {
      Cell cell = null;
      int startDirection = (int)(Math.random() * 6),
        direction = (startDirection + 1) % 6;
      while (cell == null && direction != startDirection) {
          cell = getNeighbor (direction);

	  if (!cell.isEmpty() ||
	     (!Params.thingsCanAbut &&
	      cell.anyNeighborsFilled (exceptThing)))
	  cell = null;

	  direction = (direction + 1) % 6;
      }
      return cell;
  } // method getRandomEmptyNeighbor
  
  /**
  * Returns the Cell which is n Cells away from this Cell
  * in the given direction, or null if there isn't any Cell
  * there.  There is no interpretation for a negative n.
  */
  public Cell getNthCell (int n, int direction) {
      Cell cell = this;
      
      if (n < 0) {
	  ui.write ("There's no such thing as a negative distance");
      } else {
	  int distance = 0;
	  
	  while (distance < n && cell != null) {
	      distance++;
	      cell = cell.getNeighbor (direction);
	  }
      }
      
      return (cell);
  } // method getNthCell
  
  public boolean isFertile() {
      return fertile;
  } // isFertile()

  // Methods that set variables.
  
  /**
  * Sets the contents of the Cell to be newContents.
  */
  public void setContents (Thing newContents) {
      contents = newContents;
  } // method setContents
  
  public void setFertile(boolean value) {
      fertile = value;
      if (value)
        setBGColor(FERTILITY_COLOR);
      else
	setBGColor(null);
  } // setFertile(boolean)
  
  /**
  * Sets the neighbor of the Cell in the given cardinal
  * direction to be the given cell.  Called in World when it
  * fills itself up with Cells.
  */
  public void setNeighbor(int direction, Cell cell) {
    neighbors[direction] = cell;
  } // method setNeighbor
  
  public void setBGColor(Color c) {
      bgColor = c;
  } // setBGColor(Color)
  
  public void setOutlineColor(Color c) {
      outlineColor = c;
  } // setOutputColor(Color)
  
  private Color getBGColor() {
      if (bgColor == null)
	  return BG_COLOR;
      else
	  return bgColor;
  } // getBGColor()
  
  public void paint(Graphics g) {
      fill(g, getBGColor(), true);
      drawGrid(g);
  } // method paint(Graphics)
  
  /**
  * Clears the Cell of its contents and redraws it.
  */
  public void clear() {
    Graphics g = getGraphics();

    contents = null;

    if (!ui.animationEnabled) {
      fill (g, getBGColor(), true);
      drawGrid(g);
    }
  } // method clear
  
  /**
  * Just hides the contents.
  */
  public void hide() {
      Graphics g = getGraphics();
      
      fill(g, getBGColor(), true);
      drawGrid(g);
  } // hide()
  
  /**
  * Clears the Cell of its contents and redraws it
  * using a trace color.
  */
  public void clear (Color traceColor) {
    Graphics g = getGraphics();

    contents = null;

    fill (g, traceColor, true);
    drawGrid(g);
  } // method clear
  
  // Methods which show the state of the Cell (including the Thing
  // it contains) either graphically or in text.
  
  /**
  * Writes the current coordinates of the Critter to the UI.
  */
  public void describe () {
    ui.write ("  Cell: column " + column + ", row " + row);
  } // method describe
  
  /**
  * Draws a line from the center to one point on th circle.
  */
  public void drawRadius (int heading) {
      drawRadius (getGraphics(), heading);
  } // method drawRadius
  
  public void drawRadius (Graphics g, int heading) {
      g.setColor (Heading.COLOR);
      
      if (heading == N)
	  g.drawLine (centerX, centerY, circleX[0], circleY[0]);
	  
      else if (heading == NE)
	  g.drawLine (centerX, centerY, circleX[1], circleY[1]);
	  
      else if (heading == SE)
	  g.drawLine (centerX, centerY, circleX[2], circleY[2]);
	  
      else if (heading == S)
	  g.drawLine (centerX, centerY, circleX[3], circleY[3]);
	  
      else if (heading == SW)
	  g.drawLine (centerX, centerY, circleX[4], circleY[4]);
	  
      else
	  g.drawLine (centerX, centerY, circleX[5], circleY[5]);
  } // method drawRadius

  /**
  * Five methods which fill the Cell hexagon with a color.
  */
  public void fill(Color c) {
      fill (getGraphics(), c);
  } // method fill

  public void fill(Graphics graphics, Color c) {
  	graphics.setColor (c);
  	graphics.fillPolygon (hex);
  } // method fill

  public void fill(Graphics g) {
  	g.fillPolygon (hex);
  } // method fill
  
  public void fill(Graphics g, Color c, boolean grid) {
      fill (g, c);
      if (grid)
	  drawGrid(g);
  } // method fill
  
  private void drawGrid(Graphics g) {
      if (outlineColor != null)
	  outline(g, outlineColor);
      else {
//          outline(g, World.GRID_COLOR);
	  drawCenter (g, World.GRID_COLOR);
      }
  } // method drawGrid(Graphics)
  
  public void fill(Color c, boolean grid) {
      fill (getGraphics(), c, grid);
  } // method fill
  
  public void circleOutline(Color c) {
      circleOutline(getGraphics(), c);
  } // circleOutline(Color)
  
  public void circleOutline(Graphics g, Color c) {
      g.setColor(c);
      g.drawOval(centerX - circleRad, centerY - circleRad,
		 circleDiam, circleDiam);
  } // circleOutline(Graphics, Color)
  
  public void circleFill(Color c) {
      circleFill(getGraphics(), c);
  } // method circleFill
  
  public void circleFill(Graphics g, Color c) {
      g.setColor(c);
      g.fillOval(centerX - circleRad, centerY - circleRad,
		 circleDiam, circleDiam);
  } // method circleFill
  
  public void smallCircleFill(Color c) {
      smallCircleFill(getGraphics(), c);
  } // smallCircleFill(Color)
  
  public void smallCircleFill(Graphics g, Color c) {
      g.setColor(c);
      g.fillOval(centerX - smallCircleRad, centerY - smallCircleRad,
		 smallCircleDiam, smallCircleDiam);
  } // smallCircleFill(Graphics, Color)
  
  public void arcFill(int sixth) {
      arcFill(BG_COLOR, sixth);
  }
  
  public void arcFill(Color c, int sixth) {
      arcFill(getGraphics(), c, sixth);
  }
  
  public void arcFill(Graphics g, int sixth) {
      arcFill(g, BG_COLOR, sixth);
  }
  
  public void arcFill(Graphics g, Color c, int sixth) {
      g.setColor(c);
      g.fillArc(centerX - circleRad, centerY - circleRad,
		circleDiam, circleDiam, ((7 - sixth) % 6) * 60, 60);
  }
  
  /**
  * Three methods which draw the outline of the Cell hexagon
  * in a color.
  */
  public void outline (Color c) {
      outline (getGraphics(), c);
  } // method outline
  
  public void outline (Graphics graphics) {
      graphics.drawPolygon (hex);
  } // method outline

  public void outline (Graphics graphics, Color c) {
      graphics.setColor (c);
      outline (graphics);
  } // method outline
  
  /**
  * These draw a little circle in the center of the Cell.
  */
  public void drawCenter (Graphics graphics, Color c) {
      graphics.setColor (c);
      drawCenter (graphics);
  } // method drawCenter
  
  public void drawCenter(Graphics graphics) {
      graphics.fillOval (centerX - GRID_DOT_RADIUS - 1, centerY - GRID_DOT_RADIUS - 1,
			 GRID_DOT_DIAMETER, GRID_DOT_DIAMETER);
  } // method drawCenter
  
  public void write(Graphics g, char character, Color color) {
      g.setColor(color);
      g.setFont(FONT);
      g.drawString(String.valueOf(character), xText, yText);
  } // method write(Graphics, String)
  
  public void write(char character, Color color) {
      write(getGraphics(), character, color);
  } // method write(char, Color)

  // Help methods.
  
  /**
  * Returns the Polygon that defines the Cell's hexagonal shape.
  * Calculates the X and Y coordinates of this hexagon.
  */
  private void makeHex() {
    int
      y1 = (int)(World.SQRT_THREE * radius),
      y2 = (int)(World.SQRT_THREE * 2 * radius);
      
    centerX = x + 2 * radius + 1;
    centerY = y + y1 + 1;
      
    hexX = new int[6];
    hexY = new int[6];

    hexX[0] = x + radius;
    hexX[1] = x + 3 * radius;
    hexX[2] = x + 4 * radius;
    hexX[3] = x + 3 * radius;
    hexX[4] = x + radius;
    hexX[5] = x;
      
    hexY[0] = y;
    hexY[1] = y;
    hexY[2] = y + y1;
    hexY[3] = y + y2;
    hexY[4] = y + y2;
    hexY[5] = y + y1;
      
    hex = new Polygon (hexX, hexY, 6);
  } // method makeHex
  
  /**
  * Used in drawing Critters on this Cell.
  */
  private void setCircleValues() {
    int xOffset, yOffset;
      
    circleRad = 5 * radius / 3;
    if (radius < 6) {
	circleRad--;
	circleDiam = 2 * circleRad;
    } else
      circleDiam = 2 * circleRad - 1;
  
    smallCircleRad = radius - 1;
//    circleRad / 2 + 1;
    smallCircleDiam = 2 * smallCircleRad - 1;
//    if (radius > 5)
//      smallCircleRad++;
    
    circleX = new int[6];
    circleY = new int[6];
    xOffset = (int)((circleRad - 1) * Math.cos(SIXTH_PI));
    yOffset = (int)((circleRad - 1) * Math.sin(SIXTH_PI));
    
    xText = centerX - radius / 2;
    yText = centerY + radius / 3;
    if (radius < 6)
	yText += 2;
    
    // The easy ones.
    circleX[0] = centerX;
    circleY[0] = centerY - circleRad + 1;
    circleX[3] = centerX;
    circleY[3] = centerY + circleRad - 1;
    // The others.
    circleX[1] = centerX + xOffset;
    circleY[1] = centerY - yOffset;

    circleX[2] = centerX + xOffset;
    circleY[2] = centerY + yOffset;
    
    circleX[4] = centerX - xOffset;
    circleY[4] = centerY + yOffset;

    circleX[5] = centerX - xOffset;
    circleY[5] = centerY - yOffset;
  } // method setCircleValues
  
  /**
  * Converts an int to one of the Colors possible in Cells.
  */
  public static Color intToColor (int i) {
      return COLORS[i];
  } // method intToColor
  
  public static double heading2Angle(int heading) {
      return (heading * Math.PI / 3.0);
  } // method heading2Angle

} // class Cell